import logging
import smtplib
import os
import mimetypes

from email import encoders
from email.message import Message
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from terminalinterface import *
from config import *
import minus.minus as minus

def throw(to, paths, name=None):
    t = Thrower()
    t.throw(to, paths, name=name)

class Thrower(object):
    MAX_EMAIL_SIZE = 1000000 # 1MB

    log = logging.getLogger('Thrower')

    def __init__(self):
        self._interface = TerminalInterface()
        self._config = Config()

    def throw(self, to, paths, name=None):
        if to is None or len(to) == 0:
            self._interface.new_section()

            to = [ ]

            # We need to get some recipients
            self._interface.message("""
            Before I can throw your files at a recipient, I need to know where
            to send them.
            
            I'm going to ask you for a list of recipients. I'll keep going
            until you stop giving me e-mail addresses by just pressing 'enter'
            at the prompt.""")

            while len(to) == 0:
                should_continue = True
                while should_continue:
                    to.append(self._interface.input('E-mail address to send files to'))
                    should_continue = (len(to[-1]) > 0)
                to = to[:-1]

                if len(to) == 0:
                    self._interface.error("You need to give me at least one recipient.")

        identity = {
            'name': self._config.get('user', 'name'),
            'email': self._config.get('user', 'email'),
        }

        if identity['name'] is None or identity['email'] is None:
            self._interface.new_section()
            self._interface.message("""
            I'm going to send your file by email but before I do that, I need 
            to know your name and email address""")

            if identity['name'] is None:
                identity['name'] = self._interface.input('Your name')

            if identity['email'] is None:
                identity['email'] = self._interface.input('Your e-mail address')

            self._interface.message("""
            Would you like me to remember your answers for next time? You can
            change the name and email address I use to send email later.""")
            if self._interface.input_boolean('Rememeber these values'):
                self._config.set('user', 'name', identity['name'])
                self._config.set('user', 'email', identity['email'])

        # Get a list of all the individual files to add.
        def append_dir(paths, dirpath):
            contents = [os.path.join(dirpath, x) for x in os.listdir(dirpath)]
            paths += [x for x in contents if os.path.isfile(x)]
            for subdirpath in [x for x in contents if os.path.isdir(x)]:
                append_dir(paths, subdirpath)

        filepaths = []
        filepaths += [x for x in paths if os.path.isfile(x)]
        for dirpath in [x for x in paths if os.path.isdir(x)]:
            append_dir(filepaths, dirpath)

        # Compute the total size of the files
        total_size = 0
        for path in filepaths:
            total_size += os.path.getsize(path)

        self._interface.new_section()
        self._interface.message("""
        You've asked me to throw %s file(s) with a total size of %s MB.""" % \
                (len(filepaths), total_size / 1000000.0))

        outer = MIMEMultipart()
        if name is None:
            outer['Subject'] = 'Files thrown at you'
        else:
            outer['Subject'] = 'Files thrown at you: %s' % (name,)
        outer['From'] = '%s <%s>' % (identity['name'], identity['email'])
        outer['To'] = ', '.join(to)
        outer.preamble = 'Here are some files for you'

        if(total_size < Thrower.MAX_EMAIL_SIZE):
            # Less than the maximum email size, email directly
            self._attach_files(outer, filepaths, name)
        else:
            self._share_files(outer, filepaths, name)

        try:
            # Try sending using our local SMTP server first
            server = smtplib.SMTP()
            server.sendmail(outer['From'], to, outer.as_string())
            server.quit()
        except smtplib.SMTPServerDisconnected:
            # If that failed, log into a server
            self._interface.message("""Attempting to send via GMail. Enter
            your GMail address, for example 'steve@gmail.com', and your
            password.""")

            usernm = self._interface.input('Username')
            passwd = self._interface.input('Password', no_echo=True)

            # Add the '@gmail.com' part if omitted.
            if '@' not in usernm:
                usernm += '@gmail.com'

            server = smtplib.SMTP('smtp.gmail.com', 587)
            server.starttls()
            server.login(usernm, passwd)
            server.sendmail(outer['From'], to, outer.as_string())
            server.quit()


    def _share_files(self, outer, filepaths, name):
        gallery = minus.CreateGallery()

        if name is not None:
            gallery.SaveGallery(name)

        self._interface.new_section()
        self._interface.message(\
            'Uploading files to http://min.us/m%s...' % (gallery.reader_id,))

        item_map = { }
        for path in filepaths:
            self._interface.message('Uploading %s...' % (os.path.basename(path),))
            item = minus.UploadItem(path, gallery,
                    desiredName=os.path.basename(path))
            item_map[item.id] = os.path.basename(path)

        msg_str = ''
        msg_str += "I've shared some files with you. They are viewable as a "
        msg_str += "gallery at the following link:\n\n - http://min.us/m%s\n\n" %\
                (gallery.reader_id,)
        msg_str += "The individual files can be downloaded from the following "
        msg_str += "links:\n\n"

        for item, name in item_map.items():
            msg_str += ' - http://i.min.us/j%s%s %s\n' % \
                    (item, os.path.splitext(name)[1], name)

        msg = MIMEText(msg_str)
        msg.add_header('Format', 'Flowed')
        outer.attach(msg)

    def _attach_files(self, outer, filepaths, name):
        def add_file_to_outer(path):
            if not os.path.isfile(path):
                return

            # Guess the content type based on the file's extension.  Encoding
            # will be ignored, although we should check for simple things like
            # gzip'd or compressed files.
            ctype, encoding = mimetypes.guess_type(path)
            if ctype is None or encoding is not None:
                # No guess could be made, or the file is encoded (compressed), so
                # use a generic bag-of-bits type.
                ctype = 'application/octet-stream'
            maintype, subtype = ctype.split('/', 1)

            if maintype == 'image':
                fp = open(path, 'rb')
                msg = MIMEImage(fp.read(), _subtype=subtype)
                fp.close()
            elif maintype == 'audio':
                fp = open(path, 'rb')
                msg = MIMEAudio(fp.read(), _subtype=subtype)
                fp.close()
            elif maintype == 'text':
                # We do this to catch cases where text files have
                # an encoding we can't guess correctly.
                try:
                    fp = open(path, 'r')
                    msg = MIMEText(fp.read(), _subtype=subtype)
                    fp.close()
                except UnicodeDecodeError:
                    fp = open(path, 'rb')
                    msg = MIMEBase(maintype, subtype)
                    msg.set_payload(fp.read())
                    encoders.encode_base64(msg)
                    fp.close()
            else:
                fp = open(path, 'rb')
                msg = MIMEBase(maintype, subtype)
                msg.set_payload(fp.read())
                fp.close()
                # Encode the payload using Base64
                encoders.encode_base64(msg)

            # Set the filename parameter
            msg.add_header('Content-Disposition', 'attachment',
                    filename=os.path.basename(path))
            outer.attach(msg)

        outer.attach(MIMEText("Here are some files I've thrown at you."))
        
        for path in filepaths:
            self._interface.literal_message('.')
            add_file_to_outer(path)




import json
import os

class Config(object):
    _instance = None
    _config_path = os.path.expanduser('~/.config/throw/throw.json')

    # Implement the singleton pattern
    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(Config, cls).\
                __new__(cls, *args, **kwargs)

        return cls._instance

    def __init__(self):
        # Attempt to load the config file
        try:
            self._config_dict = json.load(open(Config._config_path, 'r'))
        except IOError:
            self._config_dict = { }

    def _sync(self):
        try:
            fp = open(Config._config_path, 'w')
        except IOError:
            os.makedirs(os.path.dirname(Config._config_path))
            fp = open(Config._config_path, 'w')
        json.dump(self._config_dict, fp, indent=4)

    def exists(self, section, option):
        if section not in self._config_dict:
            return False
        if option not in self._config_dict[section]:
            return False
        return True
    
    def get(self, section, option, fallback=None):
        if not self.exists(section, option):
            return fallback
        return self._config_dict[section][option]
    
    def set(self, section, option, value):
        if section not in self._config_dict:
            self._config_dict[section] = { }

        self._config_dict[section][option] = value
        self._sync()



import sys
import logging
import formatter

try:
    import curses
    _has_curses = True
except:
    _has_curses = False

class TerminalInterface(object):
    _instance = None
    _log = logging.getLogger('Throw.TerminalInterface')

    class DumbBackend(object):
        def __init__(self, output_stream):
            self._width = 72
            self._output = output_stream
            self._writer = formatter.DumbWriter(output_stream, maxcol=self._width)

        def horiz_rule(self):
            self._writer.send_literal_data('-' * self._width)
            self._writer.send_paragraph(1)

        def literal_message(self, message_str):
            self._writer.send_literal_data(message_str)
            self._writer.send_paragraph(2)

        def message(self, message_str):
            for paragraph in message_str.split('\n\n'):
                self._writer.send_flowing_data(paragraph.strip())
                self._writer.send_paragraph(2)

        def error(self, message_str):
            self.message(message_str)

        def input(self, prompt = '', no_echo = False):
            if no_echo:
                import getpass
                return getpass.getpass(prompt)
            else:
                # This magic is to support Python 3 as well as Python 2.
                try:
                    return raw_input(prompt)
                except NameError:
                    return input(prompt)

    class CursesBackend(DumbBackend):
        # Some of this class is taken from 
        # http://code.activestate.com/recipes/475116-using-terminfo-for-portable-color-output-cursor-co/
        
        _COLORS = "BLACK BLUE GREEN CYAN RED MAGENTA YELLOW WHITE".split()
        _ANSICOLORS = "BLACK RED GREEN YELLOW BLUE MAGENTA CYAN WHITE".split()

        def __init__(self, output_stream):
            TerminalInterface.DumbBackend.__init__(self, output_stream)

            curses.setupterm(fd = output_stream.fileno())

            # Colors
            self.RESET_FG_BG = curses.tigetstr('op').decode('ascii')

            set_fg = curses.tigetstr('setf').decode('ascii')
            if set_fg:
                for i,color in zip(range(len(self._COLORS)), self._COLORS):
                    setattr(self, color, curses.tparm(set_fg, i).decode('ascii') or '')
            set_fg_ansi = curses.tigetstr('setaf').decode('ascii')
            if set_fg_ansi:
                for i,color in zip(range(len(self._ANSICOLORS)), self._ANSICOLORS):
                    setattr(self, color, curses.tparm(set_fg_ansi, i).decode('ascii') or '')
            set_bg = curses.tigetstr('setb').decode('ascii')
            if set_bg:
                for i,color in zip(range(len(self._COLORS)), self._COLORS):
                    setattr(self, 'BG_'+color, curses.tparm(set_bg, i).decode('ascii') or '')
            set_bg_ansi = curses.tigetstr('setab').decode('ascii')
            if set_bg_ansi:
                for i,color in zip(range(len(self._ANSICOLORS)), self._ANSICOLORS):
                    setattr(self, 'BG_'+color, curses.tparm(set_bg_ansi, i).decode('ascii') or '')

        def message(self, message_str):
            # Update the writer with the current terminal width
            self._width = min(curses.tigetnum('cols'), 72)
            self._writer = formatter.DumbWriter(self._output, maxcol=self._width)

            TerminalInterface.DumbBackend.message(self, message_str)

        def error(self, message_str):
            # Update the writer with the current terminal width
            self._width = min(curses.tigetnum('cols'), 72)
            self._writer = formatter.DumbWriter(self._output, maxcol=self._width)

            self._writer.send_literal_data(self.RED)
            TerminalInterface.DumbBackend.message(self, message_str)
            self._writer.send_literal_data(self.RESET_FG_BG)

        def input(self, prompt, *args, **kwargs):
            return TerminalInterface.DumbBackend.input(self,
                self.GREEN + prompt + self.RESET_FG_BG, *args, **kwargs)

    # Implement the singleton pattern
    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(TerminalInterface, cls).__new__(cls, *args, **kwargs)

        return cls._instance

    def __init__(self, stream=sys.stdout):
        # Try to import curses and setup a terminal if our output is a TTY.
        if stream.isatty() and _has_curses:
            try:
                self._backend = TerminalInterface.CursesBackend(stream)
            except curses.error:
                self._backend = TerminalInterface.DumbBackend(stream)
        else:
            # By default, use a dumb terminal backend
            self._backend = TerminalInterface.DumbBackend(stream)

    def new_section(self):
        self._backend.horiz_rule()

    def message(self, message):
        self._backend.message(message)

    def error(self, message):
        self._backend.error(message)

    def literal_message(self, message):
        self._backend.literal_message(message)

    def input(self, prompt, *args, **kwargs):
        return self._backend.input(prompt=prompt + ': ', *args, **kwargs)

    def input_boolean(self, prompt):
        while True:
            input_val = self._backend.input(prompt + ' (YES/NO): ')

            if len(input_val) > 0:
                if input_val[0] == 'y' or input_val[0] == 'Y':
                    return True
                elif input_val[0] == 'n' or input_val[0] == 'N':
                    return False
            
            self.message("I'm afraid I didn't understand that: please type 'YES' or 'NO'.")



import argparse
import logging

from thrower import *

def main():
    parser = argparse.ArgumentParser(description='Simply share a file.')
    parser.add_argument('paths', metavar='PATH', type=str, nargs='*',
        help='the name of a file or directory to share.')
    parser.add_argument('-v', '--verbose', dest='verbose',
        action='store_true',
        help='print verbose loggging information.')
    parser.add_argument('-t', '--to', dest='to', metavar='RECIPIENT', 
        action='append',
        help='a recipient for the file(s).')
    parser.add_argument('-n', '--name', dest='name', metavar='NAME', 
        action='store', default=None,
        help='a descriptive name for this set of files.')

    args = parser.parse_args()

    if(args.verbose):
        logging.basicConfig(level=logging.INFO)

    throw(args.to, args.paths, args.name)

if __name__ == '__main__':
    main()



from setuptools import setup, find_packages

import version

setup(
    name = "throw",
    version = version.get_git_version(),
    packages = find_packages(),

    entry_points = {
        'console_scripts': ['throw=throw.throw:main'],
    },

    install_requires = ['argparse'],

    # metadata for upload to PyPI
    author = "Rich Wareham",
    author_email = "rjw57@cam.ac.uk",
    description = "Simply share files from the command line",
    license = "Apache 2.0",
    keywords = "file sharing throw email",
    url = "http://github.com/rjw57/throw",

    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Topic :: Communications',
        'Topic :: Communications :: Email',
        'Topic :: Communications :: File Sharing',
        'Topic :: Utilities',

        # make sure to use :: Python *and* :: Python :: 3 so
        # that pypi can list the package on the python 3 page
        'Programming Language :: Python',
        'Programming Language :: Python :: 3'
    ],
)



try:
    import httplib
except ImportError:
    import http.client as httplib

try:
    import urllib
    import urllib2
except ImportError:
    import urllib.request
    import urllib.parse
    import urllib.error

try:
    import cookielib
except ImportError:
    import http.cookiejar as cookielib

import mimetypes
import json


class User(object):
    def __init__(self, username, password1=None):
        self.username = username
        self.password1 = password1

    def SignIn(self):
        """Signs in the User.
        User.password1 attribute must be set to the passsword before calling this method.
        Otherwise, an Exception will be raised."""
        if not self.password1:
            raise Exception("There is no password set. Set the User.password1 attribute to the password before calling this method.")
        else:
            url = "http://min.us/api/SignIn"
            params = {"username":self.username, "password1":self.password1}

            self.cj = cookielib.CookieJar()
            try:
                self.opener = urllib2.build_opener(\
                        urllib2.HTTPCookieProcessor(self.cj))
                self.login_data = urllib.urlencode(params)
            except (NameError, AttributeError):
                self.opener = urllib.request.build_opener(\
                        urllib2.HTTPCookieProcessor(self.cj))
                self.login_data = urllib.parse.urlencode(params)
            reponse = self.opener.open(url, self.login_data)
            
            parsed = json.loads(reponse.readlines()[0])
            if parsed["success"] == False:
                raise Exception("Sign in failed.")
            return parsed


    def SignOut(self):
        """Signs the User out of min.us."""
        url = "http://min.us/api/SignOut"
        response = self.opener.open(url)

    def MyGalleries(self, returnObject=False):
        """Returns a list of Gallery objects if returnObject=True.
        If returnObject=False, it will return the parsed response
        from the server as a dict.

        Both the Gallery and dict contain the same information,
        but both are availble for diversity.

        NOTE: If the editor_id is not available, it is set to 'Unavailable'
              because that's what the API json value is."""

        # TODO
        # Check if the API returns Unavailable for the editor_id
        

        url = "http://min.us/api/MyGalleries.json"

        response = self.opener.open(url)
        parsed = _parseResponse  

        if returnObject == True:
            galleryList = []
            for gallery in parsed["galleries"]:
                galleryList.append(Gallery(gallery["reader_id"], editor_id=gallery["editor_id"], \
                                           name=gallery["name"], last_visit=gallery["last_visit"], \
                                           item_count=gallery["item_count"], \
                                           clicks=gallery["clicks"]))
        elif returnObject == False:
            return parsed

    def Galleries(self):
        """To be used in the future if other users' Galleries are available through the API.
        
        This method currently does nothing."""
        pass

class Gallery(object):
    """Represents a min.us Gallery"""
    def __init__(self, reader_id, editor_id=None, name=None, last_visit=None, \
                 item_count=None, clicks=None):
        self.editor_id = editor_id
        self.reader_id = reader_id
        self.name = name

        # attributes to be used by the User.MyGalleries() method
        self.last_visit = last_visit
        self.item_count = item_count
        self.clicks = clicks
        
        # items is a list of direct links to multiple items in the gallery.
        # Use GetItems() to update this attribute.
        self.items = None

    def GetItems(self):
        """Updates self.name and self.items and returns (self.name, self.items)"""
        url = 'http://min.us/api/GetItems/' + 'm' + self.reader_id
        response = _doget(url)

        self.name = response["GALLERY_TITLE"]

        # To get the item id, we have to take the file name from the URL
        # We also need to get rid of any file extension if there is any
        self.items = [a[16:].split('.')[0] for a in response["ITEMS_GALLERY"]]

        return (self.name, self.items)
       
        
    def SaveGallery(self, name=None, items=None):
        """Use this to update the gallery name or change sort order.
        Specify which attribute (name or items or both) you want to change."""

        url = 'http://min.us/api/SaveGallery'
        
        if not name:
            if not self.name:
                name = self.GetItems()[0]
            if self.name:
                name = self.name

        if not items:
            if not self.items:
                items = self.GetItems()[1]
            elif self.items:
                items = self.items
        
        params = {"name": name, "id":self.editor_id, "items":items}

        try:        
            response = _dopost(url, params)
        except:
            pass
        else:
            self.name = name
            self.items = items
                    



class Item(object):
    """Represents a min.us item."""
    def __init__(self, id, height=None, width=None, filesize=None):
        self.id = id
        self.height = height
        self.width = width
        self.filesize = filesize


def CreateGallery():
    """Creates a Gallery on the server. Returns a Gallery object with the editor_id and reader_id."""
    url = 'http://min.us/api/CreateGallery'

    response = _dopost(url)

    _editor_id = response["editor_id"]
    _reader_id = response["reader_id"]
    
    return Gallery(_reader_id, editor_id=_editor_id)
    

def UploadItem(filename, gallery, desiredName=None):
    """filename is the full file location and name of the file
    WARNING: If your desiredName doesn't have a proper file extension (SHOULD be the same as the filename)
             it'll still upload, but you won't be able to download it or view it online.
             You can edit the name later to add the extension, but be careful, because it seems like you
             can't change the extension after that."""

    url = 'http://min.us/api/UploadItem?'   # Must have the ? because urlencode doesn't add that on itself

    if desiredName:
        name = desiredName
    else:
        name = filename

    params = {"editor_id":gallery.editor_id, "filename":name}
    
    with open(filename, 'rb') as f:
        itemData = f.read()
        
    response = _dopost(url, params=params, payload=itemData)

    _id = response["id"]
    _height = response["height"]
    _width = response["width"]
    _filesize = response["filesize"]

    return Item(_id, height=_height, width=_width, filesize=_filesize)    

def _doget(url):
    try:
        response = urllib2.urlopen(url)
    except NameError:
        response = urllib.request.urlopen(url)
    return _parseResponse(response)


def _dopost(url, params=None, payload=None):
    if params:
        try:
            encoded = urllib.urlencode(params)
        except AttributeError:
            encoded = urllib.parse.urlencode(params)
    else:
        encoded = ''

    url = str(url + encoded)

    if payload is None:
        payload = ''

    try:
        response = urllib2.urlopen(url, payload)
    except NameError:
        response = urllib.request.urlopen(url, payload)

    return _parseResponse(response) 

def _parseResponse(response):
    # response.readlines() is a list of many parts of the json.
    return json.loads(''.join([x.decode() for x in response.readlines()]))

